package Mail::SpamAssassin::CmdManage;

use strict;

use Errno;
use POSIX;

use Text::Iconv;
use MIME::Base64 ();
use MIME::QuotedPrint ();

use Mail::SpamAssassin;
use Mail::SpamAssassin::ArchiveIterator;
use Mail::SpamAssassin::Message;
use Mail::SpamAssassin::PerMsgLearner;

use Getopt::Long;
use Pod::Usage;


use constant BIG_BYTES => 256*1024;	# 256k is a big email

## Pathes definition
use constant PID_DIR  => '/var/run/spamd'; ## Pidfiles directory
use constant SOCK_DIR => '/tmp'; ## Named pipe directory
use constant PSA_CFG  => '/etc/psa/psa.conf'; ## PSA configuration file

sub SPAMC() { ## Path to Spamassassin's daemon client
	# In order to make a simultaneouth enabling/disabling of SpamAssassin
	# settings server-wide we use a special mutable wrapper
	return $main::PRODUCT_ROOT_D . '/bin/psa-spamc';
}
sub SERVER_CONFIG_DIR() { ## Server's config directory path
	return $main::LOCAL_RULES_DIR;
}
sub SERVER_CONFIG_FILE() { ## Server's config file path
	return $main::LOCAL_RULES_DIR . '/local.cf';
}

sub QMAIL_LOCAL { ## Path to qmail-local binary
	return $main::QMAIL_DIR . '/bin/qmail-local';
}

## Options definition
use constant OPTIONS => {
	'sensitivity'          => sub {
		return { 'required_score' => shift }
	},

	'modify-header'        => sub {
		return { 'rewrite_header' => 'subject ' . shift }
	},

	'not-modify-header'    => sub {
		return { 'rewrite_header' => undef }
	},

	'add-to-whitelist'     => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'whitelist_from'} || '');
		push @list, split(/,\s*/, $new); @list = map $_, keys %{ { map { $_, 1 } @list } };

		return { 'whitelist_from' => join(' ', @list) }
	},

	'add-to-blacklist'     => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'blacklist_from'} || '');
		push @list, split(/,\s*/, $new); @list = map $_, keys %{ { map { $_, 1 } @list } };

		return { 'blacklist_from' => join(' ', @list) }
	},

	'add-to-unwhitelist'   => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'unwhitelist_from'} || '');
		push @list, split(/,\s*/, $new); @list = map $_, keys %{ { map { $_, 1 } @list } };

		return { 'unwhitelist_from' => join(' ', @list) }
	},

	'add-to-unblacklist'   => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'unblacklist_from'} || '');
		push @list, split(/,\s*/, $new); @list = map $_, keys %{ { map { $_, 1 } @list } };

		return { 'unblacklist_from' => join(' ', @list) }
	},

	'del-from-whitelist'   => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'whitelist_from'} || '');
		my @del_list = split(/,\s*/, $new);
		my $i = 0; while ($i < @list) {
			if (grep $list[$i] eq $_, @del_list) { splice(@list, $i, 1) }
			else { $i++ }
		}

		my $str = join(' ', @list);
		return { 'whitelist_from' => length($str) ? $str : undef }
	},

	'del-from-blacklist'   => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'blacklist_from'} || '');
		my @del_list = split(/,\s*/, $new);
		my $i = 0; while ($i < @list) {
			if (grep $list[$i] eq $_, @del_list) { splice(@list, $i, 1) }
			else { $i++ }
		}

		my $str = join(' ', @list);
		return { 'blacklist_from' => length($str) ? $str : undef }
	},

	'del-from-unwhitelist' => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'unwhitelist_from'} || '');
		my @del_list = split(/,\s*/, $new);
		my $i = 0; while ($i < @list) {
			if (grep $list[$i] eq $_, @del_list) { splice(@list, $i, 1) }
			else { $i++ }
		}

		my $str = join(' ', @list);
		return { 'unwhitelist_from' => length($str) ? $str : undef }
	},

	'del-from-unblacklist' => sub {
		my $new = shift;
		my $old = shift;

		my @list = split(/\s+/, $old->{'blacklist_from'} || '');
		my @del_list = split(/,\s*/, $new);
		my $i = 0; while ($i < @list) {
			if (grep $list[$i] eq $_, @del_list) { splice(@list, $i, 1) }
			else { $i++ }
		}

		my $str = join(' ', @list);
		return { 'unblacklist_from' => length($str) ? $str : undef }
	},
};

## User name which the POP3 server running under
use constant POPUSER    => 'popuser';
## Apache user
use constant APACHEUSER => 'psaadm';

## Default charset for output subject line
use constant DEFAULT_CHARSET => 'utf8';

## Command line argumets
use constant COMMANDS => [ qw/start stop restart status/ ];

## Exit codes
use constant EX_OK           =>  0; ## normal exit
use constant EX_STOPPED      =>  1; ## daemon is not running
use constant EX_USAGE        => 64; ## command line usage error
use constant EX_DATAERR      => 65; ## data format error
use constant EX_NOINPUT      => 66; ## cannot open input
use constant EX_NOUSER       => 67; ## addressee unknown
use constant EX_NOHOST       => 68; ## host name unknown
use constant EX_UNAVAILABLE  => 69; ## service unavailable
use constant EX_SOFTWARE     => 70; ## internal software error
use constant EX_OSERR        => 71; ## system error (e.g., can't fork)
use constant EX_OSFILE       => 72; ## critical OS file missing
use constant EX_CANTCREAT    => 73; ## can't create (user) output file
use constant EX_IOERR        => 74; ## input/output error
use constant EX_TEMPFAIL     => 75; ## temp failure; user is invited to retry
use constant EX_PROTOCOL     => 76; ## remote error in protocol
use constant EX_NOPERM       => 77; ## permission denied
use constant EX_CONFIG       => 78; ## configuration erroruse constant EX_USAGE => 64;

## Translator errno-to-ex
use constant ERRNO2EX => {
	EACCES() => EX_NOPERM(),
	EPERM()  => EX_NOPERM(),
	ENOENT() => EX_OSERR(),
};

## Global variables
use vars qw/
	%opts %cmds $full_daemon_pid $light_daemon_pid $psa_config
	$spamtest $messagecount $learnedcount $isspam $forget
	$max_children
/;

###########################################################################
## Main access point
## >> class
## << none
sub cmdline_run($) {
	my $class  = shift;
	my $result = EX_OK();

	## First, check for we are running under right user
	$result = $class->user_check_do();
	return $result if $result;


	## Parse command line
	$result = $class->cmdline_parse();
	return $result if $result;


	## Process options
	$result = $class->bayes_train() if $opts{'bayes'};
	return $result if $result;

	$result = $class->list_mailbox() if $opts{'list-mailbox'};
	return $result if $result;

	$result = $class->change_mail_prefs()
		if $opts{'enable-mailname'} || $opts{'disable-mailname'} || $opts{'action'};
	return $result if $result;

	$result = $class->update_configuration(1)
		if grep $opts{$_}, keys %{ OPTIONS() };
	return $result if $result;

	## Process commands
	$result = $class->start_daemon(1) if $cmds{start};
	return $result if $result;

	$result = $class->stop_daemon(1) if $cmds{stop};
	return $result if $result;

	$result = $class->restart_daemon(1) if $cmds{restart};
	return $result if $result;

	$result = $class->status_daemon(1) if $cmds{status};
	return $result if $result;


	return EX_OK();
}

###########################################################################
## Checks for we are running under the right user
## >> class
## << exitcode
sub user_check_do($) {
	my $class = shift;

	## Not root - may be we are running by apache?
	if ($<) {
		my $uid = getpwnam(APACHEUSER());
		if ($< != $uid) { ## Not apache - warn and exit
			print "You are running script under user '" . getpwuid($<) . "' ";
			print "but 'root' or '" . APACHEUSER() . "' only allowed.\n";
			return EX_USAGE();
		}

		## Ok, we are under apache. Set real user/group to effective
		$< = $>; $( = $);
	}

	return EX_OK();
}

###########################################################################
## Bayesian train
## >> class
## << exitcode
sub bayes_train($) {
	my $class = shift;

	## Load PSA configuration file
	unless ($psa_config) {
		$psa_config = $class->config_parse(PSA_CFG());
		return EX_CONFIG() unless $psa_config;
	}

	## Get mails list
	my ($name, $domain) = split '@', $opts{mailname};
	my $path = $psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name";
	my $bayes_override_path = $path . '/.spamassassin/bayes';

	if ($opts{'clear'}) {
		for ($bayes_override_path . '_seen', $bayes_override_path . '_toks') {
			/^(.+)$/; ## For -T
			if (-e $1) {
				unless (unlink $1) {
					print "Cannot remove file '$_': $!\n";
					return EX_OSERR();
				}
			}
		}

		print "Database successfully removed.\n";
		return EX_OK();
	}

	## create the tester factory
	$spamtest = Mail::SpamAssassin->new({
		userprefs_filename  => $path . '/user_prefs',
		rules_filename      => $main::DEF_RULES_DIR,
		local_tests_only    => 1,
		dont_copy_prefs     => 1,
		PREFIX              => $main::PREFIX,
		DEF_RULES_DIR       => $main::DEF_RULES_DIR,
		LOCAL_RULES_DIR     => $main::LOCAL_RULES_DIR,
	});

	$spamtest->init(1);
	$spamtest->finish_learner();
	$spamtest->{conf}->{bayes_path} = $bayes_override_path;

	$spamtest->init_learner({
		force_expire      => 0,
		learn_to_journal  => 0, ## KORUM: Huh?
		wait_for_lock     => 1,
		caller_will_untie => 1,
	});

	## Sync the journal first if we're going to go r/w so we make sure to
	## learn everything before doing anything else.
	$spamtest->rebuild_learner_caches(); ## KORUM: Huh?

	## Run this lot in an eval block, so we can catch die's and clear up the dbs.
	my $result = eval {
		## We wan't just die. We need to clear place before
		$SIG{INT}  = \&killed;
		$SIG{TERM} = \&killed;

		## Run a cycle for every of type
		$messagecount = 0;
		$learnedcount = 0;
		for my $name (qw/ham spam forget/) {
			$isspam = 0; $forget = 0; ## Flush values

			## How we need to do
			if ($name eq 'ham') { $isspam = 0 }
			elsif ($name eq 'spam') { $isspam = 1 }
			elsif ($name eq 'forget') { $forget = 1 }

			## Ok, collect targets
			my $targets = $opts{$name} || ''; $targets =~ s/\\,/\&comma;/g;
			my @targets = split /,\s*/, $targets; s/\&comma;/,/g for @targets;
			@targets = map { ($isspam ? 'spam' : 'ham') . ":detect:$_" }
				grep -f, @targets;

			my $iter = new Mail::SpamAssassin::ArchiveIterator({
				'opt_j'   => 1,
				'opt_n'   => 1,
				'opt_all' => 1,
			});

			$iter->set_functions(\&wanted, sub { });

			eval { $iter->run(@targets) };
			if ($@ && $@ !~ /HITLIMIT/) { print "$@\n"; return EX_SOFTWARE() }
		}

		# There is an issue with filenames: if user simultaneously reads messages
		# in mailbox some messages may have other filename than was specified by
		# command ('S' suffix)
		# additional details you may get from http://bugs.plesk.ru/show_bug.cgi?id=23928
		# So, I'm disabling this error message until the next implementation of
		# SpamAssassin support in Plesk...
#		unless ($messagecount) {
#			print "No valid files specefied for train.\n";
#			return EX_NOINPUT();
#		}

		print "Learned from $learnedcount message(s) ($messagecount message(s) examined).\n";

		return EX_OK();
	};

	## Uups... error catched...
	if ($@) { print "$@\n"; $result = EX_SOFTWARE() }

	## Finita 'la comedie
	$spamtest->finish_learner();

	## Be a popuser
	if (($path . '/.spamassassin/') =~ /^(.+)$/) {
    	$path = $1;
    } else {
    	print "Cannot determine user's spamassassin settings directory.\n";
    	exit EX_DATAERR();
    }

	system('chown', ('-R', POPUSER() . '.' . POPUSER(), $path));

	return $result;
}

###########################################################################
## Lists user's mailbox
## >> class
## << exitcode
sub list_mailbox($) {
	my $class = shift;

	## Load PSA configuration file
	unless ($psa_config) {
		$psa_config = $class->config_parse(PSA_CFG());
		return EX_CONFIG() unless $psa_config;
	}

	## Get mails list
	my ($name, $domain) = split '@', $opts{mailname};
	my $path = $psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name";
	my ($news_path, $curs_path) = ($path . '/Maildir/new/', $path . '/Maildir/cur/');

	my @files_list;

	# Workaround for glob problem in perl 5.005 on FreeBSD (Bug 25408)
	if ($] <= 5.00503) {
		$news_path = $1 if $news_path =~ /^(.+)$/;
		$curs_path = $1 if $curs_path =~ /^(.+)$/;

		opendir(DIR, $news_path);
		my @news = grep { -f $news_path.'/'.$_ } readdir(DIR);
		closedir DIR;
		foreach my $filename (@news) {
			$filename = $news_path.'/'.$filename;
		}

		opendir(DIR, $curs_path);
		my @curs = grep { -f $curs_path.'/'.$_ } readdir(DIR);
		closedir DIR;
		foreach my $filename (@curs) {
			$filename = $curs_path.'/'.$filename;
		}

		@files_list = sort (@news, @curs);

	} else {
		@files_list = sort grep { -f $_ } (<$news_path*>, <$curs_path*>);
	}

	# my @files_list = sort grep { -f $_ } (<$news_path*>, <$curs_path*>);

	for (@files_list) {
		unless (open(INPUT, $_)) {
			print "Cannot open '$_': $!\n";
			return EX_NOINPUT();
		}
		next if (-s INPUT) > BIG_BYTES(); ## Skip too big messages

		## Ok, load it to memory
		my @mail = <INPUT>;
		close(INPUT);

		## Parse headers
		my $ma = Mail::SpamAssassin::Message->new({'message' => \@mail});
		chomp(my $subject = $ma->get_pristine_header('Subject'));
		chomp(my $from    = $ma->get_pristine_header('From'));
		undef $ma; ## Clean 'em all!

		## Print all 'bout mail, but first - decode all of them
		for ($subject, $from) {
			$_ = $class->decode_header($_);
			s/\t/    /g; s/[\n\r]+/ /g;
		}
		print "$_	$subject	$from\n";
	}

	return EX_OK();
}

###########################################################################
## Decodes MIME-coded header
## >> class
## >> raw header
## << clear header
sub decode_header($$) {
	my $class  = shift;
	my $header = shift;

	## Redefine charset if we need it
	my $output_charset = $opts{'output-charset'} || DEFAULT_CHARSET();

	while ($header =~ /=\?(.+?)\?(\w)\?(.+?)\?=/) {
		my $charset = $1; my $encoding_type = $2; my $encoded_body = $3;

		my $decoded_body;
	 	if ($encoding_type eq 'Q') { ## 'Q' type
			$decoded_body = MIME::QuotedPrint::decode($encoded_body);
		} else { ## Not - 'Q' - allways 'B'
			$decoded_body = MIME::Base64::decode_base64($encoded_body);
		}

		## Ok, try to convert from subject's charset to output one
		if ($charset ne $output_charset) {
			my $converter;
			eval { $converter = Text::Iconv->new($charset, $output_charset) }; ## Catch it!
			unless ($@) { ## Uups, not catched.. try it!
				my $converted;
				$converted = $converter->convert($decoded_body);

				## Show decoded string if we can. If not - show original
				$decoded_body = $converted if $converted;
			}
		}

		## All done. Replace raw to clear message
	 	$charset = "\Q$charset\E"; $encoded_body = "\Q$encoded_body\E";
	 	$header =~ s/=\?$charset\?$encoding_type\?$encoded_body\?=/$decoded_body/;
	}

	return $header;
}

###########################################################################
## Updates configuration file
## >> class
## << exitcode
sub update_configuration($) {
	my $class = shift;

	my $filename = $class->select_config_file();
	my $config = $class->config_parse($filename);
	unless ($config) {
		print "Cannot open or parse config file '$filename'\n";
		return EX_CONFIG();
	}

	my $new_entries = { };
	for (keys %{ OPTIONS() }) {
		if (exists($opts{$_})) {
			my $entrie = OPTIONS()->{$_}->($opts{$_}, $config);
			@$new_entries{keys %$entrie} = values %$entrie;
		}
	}

	unless ($class->config_update($filename, %$new_entries)) {
		print "Cannot open or parse config file '$filename'\n";
		return EX_CONFIG();
	}

	print(($opts{'mailname'} ? 'User' : 'Server') . "'s config file updated successfully.\n");
	return EX_OK();
}

###########################################################################
## Selects current configuration file using passed options and creates it
## if it is does not exists
## >> class
## << filename
sub select_config_file($) {
	my $class = shift;

	my $filename = SERVER_CONFIG_FILE();
	if ($opts{'mailname'}) {
		## Load PSA configuration file
		unless ($psa_config) {
			$psa_config = $class->config_parse(PSA_CFG());
			return EX_CONFIG() unless $psa_config;
		}

		my ($name, $domain) = split '@', $opts{mailname};
		$filename = $psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name/user_prefs";
	}

	if ($filename =~ /^(.*)$/) {
		$filename = $1;
	}

	## Check for config file exists and create it if not
	unless (-f $filename) {
		sysopen(FOUT, $filename, O_CREAT, 0600);
		close(FOUT);

		## Also, the user's config file must be owned by 'popuser'
		if ($opts{'mailname'}) {
			my ($login, $pass, $uid, $gid) = getpwnam(POPUSER())
				or do { print POPUSER() . " not in passwd file.\n"; return EX_NOUSER() };

			chown($uid, $gid, $filename);
		}
	}

	return $filename;
}

###########################################################################
## Changes virtual qmail's user '.qmail' preferences file
## >> class
## << exitcode
sub change_mail_prefs($) {
	my $class = shift;

	## Load PSA configuration file
	unless ($psa_config) {
		$psa_config = $class->config_parse(PSA_CFG());
		return EX_CONFIG() unless $psa_config;
	}

	my ($name, $domain) = split '@', $opts{mailname};
	my $filename = $psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name/.qmail";

	## Check for file exists
	unless (-f $filename) {
		print "User's preferences file '$filename' does not exists.\n";
		return EX_CONFIG();
	}

	unless (open(FIN, $filename)) {
		print "Unable to open file '$filename'\n";
		return EX_CONFIG();
	}

	my @content;
	my $maildir;	# line containing the ./Maildir/ record
	my $spamc;		# line containing the spamc call

	while (<FIN>) {
		push @content, $_;

		if ($maildir && $spamc) {next;}

		if (/^\.\/Maildir\/$/)		{ $maildir	= $_; }
		if (/(psa-)?spamc -f -u /)	{ $spamc	= $_; }
	}
	close(FIN);

	# if no ./Maildir/ found it may be specified on the same line as spamc (Plesk 7.0.0)
	unless ($maildir) {
		if (!$spamc || !$spamc =~ /\.\/Maildir\//) {
		#unless ($spamc =~ /\.\/Maildir\//) {
			# to not confuse support ([Bug 26854]  - Host365 - Plesk 5 to Plesk 7 upgrade on two servers)
			# print "No record was found about delivery to './Mailbox' directory.\n";
			return EX_OK();
		}
	}

	###
	#	Check options
	###
	if (!$spamc) {

		###
		#	Spam filtering is currently disabled
		###
		if ($opts{'disable-mailname'}) {
			print "Spam filtering for specified mailname disabled already.\n";
			return EX_OK();
		}

	} else {

		if ($spamc =~ /qmail-queue/) {

			###
			#	Spam filtering is currently enabled with action "save"
			###
			if (	$opts{'enable-mailname'} 
				&& !$opts{'disable-mailname'} 
				&& (!$opts{'action'} || ($opts{'action'} eq 'save'))) {

				print "Spam filtering for specified mailname is already enabled with mode 'save'.\n";
				return EX_OK();
			}

		} elsif ($spamc =~ /\| \/.*\/maildir \// && !$spamc =~ /exit 99/) {

			###
			#	Artefact from Plesk 7.0.0
			###

		} else {

			###
			#	Spam filtering is currently enabled with action "delete"
			###
			if (	$opts{'enable-mailname'} 
				&& !$opts{'disable-mailname'} 
				&& ($opts{'action'} eq 'delete')) {

				print "Spam filtering for specified mailname is already enabled with mode 'delete'.\n";
				return EX_OK();
			}

		}
	}

	###
	#	Create a new file entry
	###
	my $newstr;

	if ($opts{'enable-mailname'} || $opts{'action'}) {
		my $sock_path = SOCK_DIR . '/' .
				($opts{'enable-mailname'} ?
					$opts{'enable-server-configs'} ? 'spamd_full.sock' : 'spamd_light.sock' :
					$spamc =~ /-U \S+(spamd_full.sock|spamd_light.sock)/ ? $1 : 'spamd_full.sock'
				);

		if ($opts{'action'} && ($opts{'action'} eq 'delete')) {

			###
			#	Spam shouldn't be delivered so exit 99
			###
			$newstr = sprintf(
				'| %s -f -u %s -U %s -c || exit 99' . "\n",
				SPAMC(), $opts{'mailname'}, $sock_path);

		} else {

			###
			#	Mail should be delivered altered by spamassassin
			#	The trick here is reusing qmail-local to deliver mail further
			#	by the delivery chain
			###
			$newstr = sprintf(
				'| if [ -z "$SA" ]; then export SA=1; %s -f -u %s -U %s > spamcheck$$; '.
				'%s "$USER" "$HOME" "$LOCAL" "" "" "$HOST" "$SENDER" "$DEFAULT" < spamcheck$$; '.
				'retval=$?; rm -f spamcheck$$; [ $retval = 0 ] && exit 99; exit $retval; fi' . "\n", 
				SPAMC(), $opts{'mailname'}, $sock_path, 
				QMAIL_LOCAL());
		}
	}

	###
	#	Clear the user preferences file only if "spammng -e" was called, to
	#	be sure that we aren't out of sync when SA support is enabling on the mailname
	###
	if ($opts{'enable-mailname'} && !$opts{'disable-mailname'}) {

		if (($psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name/user_prefs") =~ /^(.+)$/) {
			$filename = $1;
		} else {
			print "Cannot determine where the users' settings are located.\n";
			exit EX_DATAERR();
		}

		## Ok, now we are ready to open the file
		unless (open(DATA, ">" . $filename)) {
			print "Unable to create the file '$filename': $!\n";
			return EX_OSERROR();
		}
		close(DATA); ## All Done
	}

	###
	#	Assemble the new file contents:
	#
	#	1 "./Maildir/" not presented on the separate line means transfer from Plesk 7.0.1
	#	2 Missing spamc call means enabling spam filtering for mailname
	#	3 Missing new spamc string means disabling of spam filtering for mailname
	###
	if (!$maildir) {

		###
		#	Transfer from Plesk 7.0.0
		#	Add spamc call (if any) and ./Maildir/ line
		#	Remove previous spamc call
		###
		if ($newstr) {
			unshift @content, ($newstr);
		}
		push @content, ("./Maildir/\n");

		for (@content) {
			if ($_ eq $spamc) {
				$_ = '';
			}
		}

	} elsif (!$spamc) {

		###
		#	Spam filtering is disabled currently and should be enabled
		#	Insert the spamc call on the first place in the delivery chain
		###
		unshift @content, ($newstr);

	} elsif (!$newstr) {

		###
		#	Spam filtering is enabled currently and should be disabled
		###
		for (@content) {
			if ($_ eq $spamc) {
				$_ = '';
			}
		}

	} else {

		###
		#	Just change the action of the spamfilter: save or delete
		###
		for (@content) {
			if ($_ eq $spamc) {
				$_ = $newstr;
			}
		}
	}

	$filename = $psa_config->{'QMAIL_MAILNAMES_D'} . "/$domain/$name/.qmail";
	$filename = $1 if ($filename =~ /^(.+)$/); ## Care about '-T' flag 
	## Set sticky bit on file for prevent QMail delivery
	my $sticky = 1 << 9; ## 01000
	my $mode = (stat($filename))[2];
	$mode |= $sticky;
	chmod($mode, $filename);
	sleep(1); ## Sleep a little to get time for server before process file

	## Ok, now we can change it
	unless (open(FOUT, ">$filename")) {
		print "Unable to create file '$filename'\n";

		my $mode = (stat($filename))[2];
		$mode &= ~$sticky;
		chmod($mode, $filename);

		return EX_CONFIG();
	}

	## Write content and close it
	print FOUT for @content;
	close(FOUT);
	
	## Flush sticky bit on file for restore QMail delivery
	$mode = (stat($filename))[2];
	$mode &= ~$sticky;
	chmod($mode, $filename);

	print "User's preferences file updated successfully.\n";
	return EX_OK();
}

###########################################################################
## Starts the Spamassassin's daemon
## >> class
## >> show messages
## << exitcode
sub start_daemon($) {
	my $class = shift;
	my $msgs  = shift;

	unless ($class->status_daemon()) {
		print "Server is running already.\n" if $msgs;
		return EX_OK();
	}

	## Load PSA configuration file
	unless ($psa_config) {
		$psa_config = $class->config_parse(PSA_CFG());
		return EX_CONFIG() unless $psa_config;
	}

	chdir($main::QMAIL_DIR); ## Prevent daemon died

	############################################################
	#
	#	Watch the command line. If SpamAssassin is completely
	#	disabled by administrator - modify wrapper to not call
	#	spamc at all.
	#
	############################################################

	delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
	$ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin';

	if (!$opts{'enable-server-configs'} && 
		!$opts{'enable-user-configs'} && 
		!$opts{'mailname'} && 
		!$opts{'enable-mailname'}) {

		# SpamAssassin is completely disabled by server admin: modify wrapper
		system("perl -pi -e 's/CALL_SPAMC=\"yes\"/CALL_SPAMC=\"no\"/' ".SPAMC());

	} else {

		# Some SpamAssassin is enabled somehow: allow wrapper to call spamc
		system("perl -pi -e 's/CALL_SPAMC=\"no\"/CALL_SPAMC=\"yes\"/' ".SPAMC());

	}

	## Common arguments
	my @common_args = (qw|--username=popuser --daemonize --helper-home-dir=/var/qmail|);
	if (!$max_children) {
		$max_children = $psa_config->{'SPAMASSASSIN_MAX_CHILDREN'};
		if (!$max_children) {
			$max_children = 5;
		}
	}
	push @common_args, '--max-children', $max_children;
	push @common_args, (
		qw|--create-prefs --nouser-config|,
		'--virtual-config-dir='
		. $psa_config->{'QMAIL_MAILNAMES_D'} . '/%d/%l',
	) if $opts{'enable-user-configs'};

	## Start full configured daemon
	my @args = ( @common_args, '--pidfile=' . PID_DIR . '/spamd_full.pid' );
	push @args, ('--socketpath=' . SOCK_DIR . '/spamd_full.sock');
	push @args, (qw|--siteconfigpath=/dev/null|)
		if !$opts{'enable-server-configs'};

	# Fix of Insecure $ENV{PATH}
	delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
	$ENV{'PATH'} = $main::PREFIX . '/bin';

	system($main::PREFIX . $main::SPAMD_PREFIX . '/spamd', @args);

	my $exitcode = $? >> 8;
	if ($exitcode) {
		print "Unable to start server.\n" if $msgs;
		return ERRNO2EX()->{$exitcode} || EX_OSERR();
	}

	## Start light configured daemon
	@args = ( @common_args, '--pidfile=' . PID_DIR . '/spamd_light.pid' );
	push @args, ('--socketpath=' . SOCK_DIR . '/spamd_light.sock');
	push @args, ('--siteconfigpath=/dev/null')
		if $opts{'enable-user-configs'} || !$opts{'enable-server-configs'};

	system($main::PREFIX . $main::SPAMD_PREFIX . '/spamd', @args);

	$exitcode = $? >> 8;
	if ($exitcode) {
		print "Unable to start server.\n" if $msgs;
		return ERRNO2EX()->{$exitcode} || EX_OSERR();
	}


	print "Server has been successfully started.\n" if $msgs;
	return EX_OK();
}

###########################################################################
## Stops the Spamassassin's daemon
## >> class
## >> show messages
## << exitcode
sub stop_daemon($) {
	my $class = shift;
	my $msgs  = shift;

	## First, check daemon status
	if ($class->status_daemon()) {
		print "Server is stopped already.\n" if $msgs;
		return EX_OK();
	}

	## Ok, ask server for die
	for ($full_daemon_pid, $light_daemon_pid) {
		next unless $_; ## Skip died daemons

		unless (kill TERM => $_) {
			print "Cannot kill spamd (pid $_) process: $!\n" if $msgs;
			return EX_OSERR();
		}
	
		## Check for daemon really died
		unless ($class->status_daemon()) {
			sleep(1); ## Get it time to die
			kill KILL => $_; ## DIE! DIE!! DIE!!!
		}
	}

	print "Server has been successfully stopped.\n" if $msgs;
	return EX_OK();
}

###########################################################################
## Restarts the Spamassassin's daemon
## >> class
## >> show messages
## << exitcode
sub restart_daemon($) {
	my $class = shift;
	my $msgs  = shift;

	my $result = EX_OK();
	
	my $old_full_pid;
	my $old_light_pid;
	my $daemon_status;

	if ($class->status_daemon()) {
		print "Server is not running.\n" if $msgs;
		return EX_STOPPED();
	}

	## Ok, ask server for restart
	for ($full_daemon_pid, $light_daemon_pid) {
		unless (kill HUP => $_) {
			print "Cannot restart spamd process: $!\n" if $msgs;
			return EX_OSERR();
		}
	}
	
	## TODO: BEGIN OF Workaround: in my installation daemon will dies due receiving SIGHUP
	## if spamd receives SIGHUP, it internally reloads itself, which means that it will change its pid
	## we must wait for reloading before start
	$old_full_pid = $full_daemon_pid;
	$old_light_pid = $light_daemon_pid;
	while (!$daemon_status && ($old_full_pid == $full_daemon_pid) && ($old_light_pid == $light_daemon_pid)) {
		sleep(1);
		$old_full_pid = $full_daemon_pid;
		$old_light_pid = $light_daemon_pid;
		$daemon_status = $class->status_daemon();
	}
	## END OF Workaround

	$class->start_daemon(1) if $class->status_daemon();

	print "Server has been successfully restarted.\n" if $msgs;
	return EX_OK();
}

###########################################################################
## Status the Spamassassin's daemon
## >> class
## >> messages - output status messages
## << EX_OK for running and EX_STOPPED if else
sub status_daemon($;$) {
	my $class = shift;
	my $msgs  = shift;

	## Process full configured daemon
	my $pidfile;
	for ($full_daemon_pid, $light_daemon_pid) {
		unless ($pidfile) { $pidfile = PID_DIR . '/spamd_full.pid' }
		else { $pidfile = PID_DIR . '/spamd_light.pid' }

		## Read pidfile
		unless (open(PIDFILE, $pidfile)) {
			# This message appears when spamd simply not running (#25125)
			# warn "spammng: Pidfile does not exists or cannot be opened.\n" if $msgs;
			return EX_STOPPED();
		}

		chomp($_ = <PIDFILE>);
		close(PIDFILE);

		## Check for a number
		unless (/(\d+)/ && $1) {
			warn "spammng: Pidfile exists but does not contains valid PID.\n" if $msgs;
			return EX_STOPPED();
		}
		$_ = $1;

		## Check for proccess is alive
		unless (kill 0, $_) {
			warn "spammng: PID file exists but process (pid $_) is not running.\n" if $msgs;
			return EX_STOPPED();
		}

	}

	#warn "spamd (pid $full_daemon_pid, $light_daemon_pid) is running...\n" if $msgs;
	print "is running\n" if $msgs; ## For 'clever' PSA
	return EX_OK();
}

###########################################################################
## Parse command line arguments
## Method will exits on error
## >> class
## << none
sub cmdline_parse($) {
	my $class = shift;

	Getopt::Long::Configure(qw(no_ignore_case));

	do { print "Command line usage invalid.\n"; return EX_USAGE() }
		unless GetOptions(\%opts,
			'enable-server-configs|c',
			'enable-user-configs|C',
			'sensitivity|s=i',
			'modify-header|m=s',
			'not-modify-header|M',
			'add-to-whitelist|a=s',
			'add-to-blacklist|A=s',
			'del-from-whitelist|d=s',
			'del-from-blacklist|D=s',
			'add-to-unwhitelist|u=s',
			'add-to-unblacklist|U=s',
			'del-from-unwhitelist|l=s',
			'del-from-unblacklist|L=s',
			'enable-mailname|e',
			'disable-mailname|E',
			'action|z=s',
			'bayes|b',
			'spam|p=s',
			'ham|h=s',
			'forget|f=s',
			'clear|r',
			'mailname|n=s',
			'list-mailbox|x',
			'output-charset|o=s',
			'help|?',
		);

	## Ok, options has been parsed. Let's get commands
	for my $cmd (@{ COMMANDS() }) {
		my $i = 0;
		for my $arg (@ARGV) {
			last if $cmd eq $arg;
			$i++;
		}

		$cmds{$cmd} = 1 unless $i > $#ARGV;
		splice(@ARGV, $i, 1);
	}

	## Define on-cmd line parse error method
	my $cmd_error = sub {
		my @lines = @_;

		push @lines, 'For usage information start utility with \'--help\' option.';
		chomp(@lines); my $msg = join "\n", @lines; $msg .= "\n"; print $msg;

		return EX_USAGE();
	};

	## Ok, commands parsed. All of another - is a trash
	if (@ARGV) {
		my $s = $#ARGV ? 's' : '';
		return $cmd_error->("Unknown option$s/argument$s '@ARGV' specified.");
	}

	## Help needed
	return usage("For more information read the manual page", EX_OK())
		if $opts{'help'} || !(%opts || %cmds);

	return $cmd_error->("Too many arguments specified.")
		if keys(%cmds) > 1;


	## Command line parsed. Let's check dependencies and cross-races
	return $cmd_error->("Specified mailname is invalid.")
		if $opts{'mailname'} && $opts{'mailname'} !~ /^[\w0-9][\w_0-9.-]*\@([\w0-9][\w_0-9.-]+)?[\w0-9]$/;

	return $cmd_error->("'--enable-user-configs' specified without 'start' command.")
		if $opts{'enable-user-configs'} && !($cmds{'start'} || $cmds{'restart'});

	return $cmd_error->("'--bayes' specified without any train option.")
		if $opts{'bayes'} && !$opts{'spam'} && !$opts{'ham'} && !$opts{'forget'} && !$opts{'clear'};

	return $cmd_error->("'--enable-mailname' specified without '--mailname' option")
		if $opts{'enable-mailname'} && !$opts{'mailname'};

	return $cmd_error->("'--disable-mailname' specified without '--mailname' option")
		if $opts{'disable-mailname'} && !$opts{'mailname'};

	return $cmd_error->("'--enable-server-configs' specified without '--enable-mailname' option or 'start' argument.")
		if $opts{'enable-server-configs'} && !($opts{'enable-mailname'} || $cmds{'start'} || $cmds{'restart'});

	return $cmd_error->("'--list-mailbox' specified without '--mailname' option")
		if $opts{'list-mailbox'} && !$opts{'mailname'};

	return $cmd_error->("'--output-charset' specified without '--list-mailbox' option")
		if $opts{'output-charset'} && !$opts{'list-mailbox'};


	return EX_OK();
}

###########################################################################
## Updates Unix-like config file
## >> class
## >> filename
## >> hash of config entries for update (undef value means 'remove entrie')
## << 1 on success or undef on error
sub config_update($$) {
	my $class    = shift;
	my $filename = shift;
	my %entries  = @_;

	## Try to open file
	unless (open(CONF, $filename)) {
		print "Unable to open file '$filename': $!\n";
		return undef;
	}
	## Load all of config
	my @config = <CONF>;
	close(CONF); ## Done loading

	## Try to create file
	$filename =~ /^(.+)$/; $filename = $1; ## care about -T option
	unless (open(CONF, ">$filename")) {
		print "Unable to create file '$filename': $!\n";
		return undef;
	}

	## Ok, put all of config file but replace a new entries
	for (@config) {
		next if /^\s*#/; ## skip comments

		next unless /^(\S+)\s+(.+)$/; ## parse key-value pair

		## this entrie must be updated
		if (defined $entries{$1}) {
			$_ = "$1\t$entries{$1}\n";
			delete $entries{$1};
		}
		if (exists($entries{$1}) && !defined($entries{$1})) {
			$_ = ''; delete $entries{$1};
		}

		print CONF; ## Ok, output entrie
	}

	## Ok, config file patched. Let's add new entries
	print CONF "$_" . ($entries{$_} ne '' ? "\t$entries{$_}\n" : '')
		for grep defined $entries{$_}, keys %entries;

	close(CONF); ## All done

	return 1;
}

###########################################################################
## Parses Unix-like config file
## >> class
## >> filename
## << hashref with config data on success or undef on error
sub config_parse($$) {
	my $class    = shift;
	my $filename = shift;
	my $result   = { };

	## Try to open file
	unless (open(CONF, $filename)) {
		print "Unable to open file '$filename': $!\n";
		return undef;
	}

	## Parse all of config
	while (<CONF>) {
		s/^\s*(.+?)\s*$/$1/;

		## Skip comment lines
		next if /^#/;
		next unless /^(\S+)\s+(.+)$/;

		$result->{$1} = $2;
	}

	close(CONF); ## Well, done

	return $result;
}

###########################################################################
## Safe on-kill method. It will handles SIGTERM and SIGKILL
## >> none
## << none
sub killed {
	$spamtest->finish_learner();
	print "interrupted\n" . EX_TEMPFAIL() . "\n";
	exit(EX_TEMPFAIL());
}

###########################################################################
## Learn for one mail message
## >> message ID
## >> message count
## >> arrayref with lines of message
sub wanted {
	my ( $class, $id, $time, $dataref ) = @_;

	my $spam = $class eq "s" ? 1 : 0;
	$messagecount++;
	my $ma = $spamtest->parse($dataref);

	if ( $ma->get_header("X-Spam-Checker-Version") ) {
		my $dataref = $spamtest->remove_spamassassin_markup($ma);
		$ma->finish();
		$ma = $spamtest->parse($dataref, 1);
	}

	my $status = $spamtest->learn( $ma, undef, $spam, $forget );
	$learnedcount++ if $status->did_learn();

	# Do cleanup ...
	$status->finish();
	undef $status;

	$ma->finish();
	undef $ma;
}

###########################################################################
## Shows usage information from the caller's POD
## >> message - The text of a message to print immediately prior to printing the program's usage message
## >> exitval - The desired exit status to pass to the exit() function. This should be an integer
sub usage($$) {
	my ($message, $exitval) = @_;
	my $ver = Mail::SpamAssassin::Version();
	print "SpamAssassin version $ver\n";
	pod2usage(-verbose => 1, -message => $message, -exitval => $exitval);
}

1;

